import { useState, useEffect } from "react";

const nameToHookName = name => "use" + name.chatAt(0).toUpperCase() + name.slice(1);

const shallowEq = (x, y) => x === y;

const mapValues = (src, fn) => Object.fromEntries(Object.entries(src).map(([key, val]) => [key, fn(val)]));

const createObservable = initial => {
  let _val = initial;
  let _listeners = [];

  return {
    _get: () => _val,
    _set: newValue => {
      const oldValue = _val;
      _val = newValue;
      _listeners
        .filter(({ equality }) => !equality(oldValue, newValue))
        .forEach(({ callback }) => callback(newValue));
    },
    _sub: (callback, equality) => {
      _listeners.push({ callback, equality });
      return () => {
        _listeners = _listeners.filter(ln => ln.callback !== callback);
      }
    }
  }
}

const createHook = obs => {
  // allow linters to pick this up as a hook
  const useObservable = (equality = shallowEq) => {
    const [val, setVal] = useState(obs._get());
    useEffect(() => obs._sub(setVal, equality), [equality]);
    return val;
  }
  return useObservable;
}

const mergeState = (store, hooks, newState) => {
  Object.entries(newState).forEach(([key, newValue]) => {
    const obs = store[key];
    if (obs) {
      obs._set(newValue);
    } else {
      const newObs = createObservable(newValue);
      store[key] = newObs;
      hooks[nameToHookName(key)] = createHook(newObs);
    }
  });
}

export const createStore = (initial, actions = {}) => {
  const store = {};
  const hooks = {};
  const dispatch = mapValues(actions, act => (...args) => mergeState(store, hooks, act(...args)));
  const selector = mapValues(store, obs => obs._get);
  mergeState(store, hooks, initial);
  return [hooks, dispatch, selector];
}